import logging

import numba
import numpy as np
from numpy.typing import NDArray

LOG = logging.getLogger(__name__)

from pypole import NDArray64

_EPSILON = 1e-50


def dipolarity_param(data_map: NDArray64, fitted_map: NDArray64) -> np.float64:
    """Calculate the dipolarity parameter of a magnetic dipole field.

    The dipolarity parameter (DP) is defined as the ratio of the rms of the
    residual (fitted_map - map) to the rms of the data map.
    DP was first introduced by [1]_.

    Args:
      data_map: magnetic field map
      fitted_map: fitted magnetic field map

    Returns:
      float: dipolarity parameter of the magnetic dipole field

    References
    ----------
    .. [1] Fu, Roger R., Eduardo A. Lima, Michael W. R. Volk, and Raisa Trubko.
    “High-Sensitivity Moment Magnetometry With the Quantum Diamond Microscope.”
    Geochemistry, Geophysics, Geosystems 21, no. 8 (2020): e2020GC009147. https://doi.org/10/ghfpqv.
    """
    residual: NDArray64 = fitted_map - data_map
    return 1 - (rms(residual) / rms(data_map))


def rms(b_map: NDArray64) -> np.float64:
    """Calculate the root mean square of a map.

    Args:
      b_map: Map to calculate the RMS of

    Returns:
        float: RMS of the map
    """
    return np.sqrt(np.mean(np.square(b_map))).astype(np.float64)


def upward_continue(
    b_map: NDArray64, distance: float, pixel_size: float, oversample: int = 2
) -> NDArray64:
    """Upward continues a map.

    This function calculates a new map that is the upward continuation of the initial map by a given distance.
    In other words, it returns a new map that looks as if it was measured at a different distance from the sample.

    Args:
      map: The map to be continued
      distance: The distance to upward continue the map in m
      pixel_size: The size of the pixel in the map in m
      oversample: The oversampling factor to use (Default value = 2)

    Returns:
        The upward continued map
    """
    ypix, xpix = b_map.shape
    new_x, new_y = xpix * oversample, ypix * oversample

    # Calculate the new pixel size
    b_map = pad_map(b_map)

    # these freq. coordinates match the fft algorithm
    x_steps = np.concatenate([np.arange(0, new_x / 2, 1), np.arange(-new_x / 2, 0, 1)])
    fx = x_steps / pixel_size / new_x
    y_steps = np.concatenate([np.arange(0, new_y / 2, 1), np.arange(-new_y / 2, 0, 1)])
    fy = y_steps / pixel_size / new_y

    fgrid_x, fgrid_y = np.meshgrid(fx + _EPSILON, fy + _EPSILON)

    kx = 2 * np.pi * fgrid_x
    ky = 2 * np.pi * fgrid_y
    k = np.sqrt(kx**2 + ky**2)

    # Calculate the filter frequency response associated with the x component
    x_filter = np.exp(-distance * k)

    # Compute FFT of the field map
    fft_map = np.fft.fft2(b_map, s=(new_y, new_x))

    # Calculate single component
    b_out = np.fft.ifft2(fft_map * x_filter)
    LOG.debug("Upward continued map by %s m", distance)

    # Crop matrices to get rid of zero padding
    return b_out[ypix : 2 * ypix, xpix : 2 * xpix].real


def pad_map(b_map: NDArray64, oversample: int = 2) -> NDArray64:
    """Pads a map with zeros.

    Args:
      map: The map to be padded

    Returns:
        The padded map
    """
    pad_size = np.array(
        ((b_map.shape[0], b_map.shape[0]), (b_map.shape[1], b_map.shape[1]))
    )
    pad_size *= oversample - 1

    return np.pad(b_map, pad_width=pad_size, mode="constant", constant_values=0)



from typing import Any, Tuple

import itertools

import numba
import numpy as np
from numpy.typing import NDArray

from pypole import NDArray64
from pypole.maps import get_grid, get_random_sources


def synthetic_map(
    n_sources: int = 100,
    pixels: tuple[int, int] = (100, 100),
    sensor_distance: float = 5e-6,
    pixel_size: float = 1e-6,
) -> NDArray64:
    """Calculate a single simple magnetic field map for a number of sources.

    Parameters
    ----------
    n_sources : int
        number of sources
    pixels : int
        number of pixels in the map
    sensor_distance : float
        distance from the sensor to the sample
    pixel_size : float
        size of each pixel in meters

    Returns
    -------
    b_map : ndarray(pixels, pixels)
        magnetic field map
    """
    # get the map grid in meters
    x_grid, y_grid = get_grid(pixels, pixel_size)

    # get the location for the sources
    locations, source_vectors = get_random_sources(n_sources)
    return calculate_map(x_grid, y_grid, locations, source_vectors, sensor_distance)


def calculate_map(
    x_grid: NDArray64,
    y_grid: NDArray64,
    locations: NDArray64,
    source_vectors: NDArray64,
    sensor_distance: float = 5e-6,
) -> NDArray[np.float_]:
    """Calculate the magnetic field map for a set of sources

    Parameters
    ----------
    x_grid : ndarray(pixel, pixel)
        x grid
    y_grid : ndarray(pixel, pixel)
        y grid
    locations : ndarray(n_sources, 3)
        x,y,z locations of sources
    source_vectors : ndarray(n_sources, 3)
        dipole moments in (x, y, z) format and Am2
    sensor_distance : float, optional
        sensor-sample distance by default 0.0
    """

    n_sources: int = locations.shape[0]
    b: NDArray64 = np.zeros((n_sources, x_grid.shape[0], x_grid.shape[1]))

    for i in range(n_sources):
        b[i, :, :] = dipole_field(
            x_grid,
            y_grid,
            locations[i, 0],
            locations[i, 1],
            locations[i, 2] + sensor_distance,
            source_vectors[i, 0],
            source_vectors[i, 1],
            source_vectors[i, 1],
        )
    return np.sum(b, axis=0)


@numba.jit(fastmath=True)
def dipole_field(
    x_grid: NDArray64,
    y_grid: NDArray64,
    x_source: NDArray64,
    y_source: NDArray64,
    z_observed: NDArray64,
    mx: float,
    my: float,
    mz: float,
) -> NDArray64:
    """Compute the field of a magnetic dipole point source

    Parameters
    ----------
    x_source: ndarray (n_sources, 1)
        x-locations of source
    y_source:  ndarray (n_sources, 1)
        y-locations of source
    z_observed: ndarray(n_sources)
        observed z distance, including the sensor height
    x_grid:  ndarray(pixel, pixel)
        grid to calculate the fields for
    y_grid: ndarray(pixel, pixel)
        grid to calculate the fields on
    mx: ndarray(n_sources,)
        x-component of vector in Am2
    my: ndarray(n_sources,)
        y-component of vector in Am2
    mz: ndarray(n_sources,)
        z-component of vector in Am2
    """
    dgridx = np.subtract(x_grid, x_source)
    dgridy = np.subtract(y_grid, y_source)

    squared_distance = np.square(dgridx) + np.square(dgridy) + np.square(z_observed)

    aux = mx * dgridx + my * dgridy + mz * z_observed
    aux /= np.sqrt(np.power(squared_distance, 5.0))
    aux = 3.0 * aux * z_observed
    return 1e-7 * (aux - mz / np.sqrt(np.power(squared_distance, 3.0)))



"""Example of code."""


def hello(name: str) -> str:
    """Just an greetings example.

    Args:
        name (str): Name to greet.

    Returns:
        str: greeting message

    Examples:
        .. code:: python

            >>> hello("Roman")
            'Hello Roman!'
    """
    return f"Hello {name}!"



import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import ImageGrid

from pypole import NDArray64, compute, fit, maps


def compare_maps(map1: NDArray64, map2: NDArray64, title: str = "") -> None:
    """Plot a comparison of the data map and the fitted map

    Args:
        map1 (NDArray64): Map 1 to compare
        map2 (NDArray64): Map 2 to compare
        title (str, optional): Title of the plot. Defaults to None.

    Returns:
        None
    """

    fig = plt.figure(figsize=(10, 3))
    fig.suptitle(title)

    res = map1 - map2

    ax = ImageGrid(
        fig,
        111,
        nrows_ncols=(1, 3),
        axes_pad=0.05,
        cbar_location="right",
        cbar_mode="edge",
        cbar_size="5%",
        cbar_pad=0.05,
    )
    vmin, vmax = np.percentile(map1, [0, 100])
    ax[0].imshow(map1, vmin=vmin, vmax=vmax, origin="lower")
    ax[0].set(title="data map", xlabel="px", ylabel="px")
    ax[1].imshow(map2, vmin=vmin, vmax=vmax, origin="lower")
    ax[1].set(title="fitted map", xlabel="px", ylabel="px")
    l = ax[2].imshow(res, vmin=vmin, vmax=vmax, origin="lower")
    ax[2].set(title="residual", xlabel="px", ylabel="px")

    plt.colorbar(l, cax=ax.cbar_axes[0], label="B [T]")
    plt.tight_layout()
    plt.show()

    print(f"residual: max: {np.max(res):.2e} std: {np.std(res):.2e} T")
    print(f"dipolarity parameter: {compute.dipolarity_param(map1, map2):.2f}")



import numba
import numpy as np
from numpy.typing import ArrayLike, NDArray

from pypole import NDArray64

"""
Converts between xyz and polar/azimuth coordinates with the convention of N along -Y-axis and +Z down.

               N, -Y
             *********
         *               *
       *                   *
      *                     *
   -X *     ↑ -Z ↓ +Z       *  +X (declination = 90°)
      *                     *
       *                   *
          *             *
             *********
               S, +Y (declination = 180°)
"""


def dim2xyz(dim: NDArray64) -> NDArray64:
    """calculates x,y,z from polar/azimuth data

    Parameters
    ----------
    dim: ndarray
        declination, inclination, magnitude

    Returns
    -------
    np.array
        of x,y,z values

    >>> dim2xyz(np.array([[0,0,1], [0,90,1]]))
    array([[ 0., -1.,  0.], [ 0.,  0., -1.]])
    """
    # convert to np array
    dim = np.atleast_2d(dim)

    # separate columns
    # convert to radians
    dec = np.deg2rad(dim[:, 0])
    inc = np.deg2rad(dim[:, 1])
    mag = dim[:, 2]

    moment_vector = [
        mag * np.sin(dec) * np.cos(inc),
        -mag * np.cos(dec) * np.cos(inc),
        mag * np.sin(-inc),
    ]
    return np.array(moment_vector).T


def xyz2dim(xyz: ArrayLike) -> NDArray64:
    """calculates polar/azimuth from x,y,z data

    Parameters
    ----------
    xyz: ndarray (3,n)
        x,y,z values of n vectors [[x1,y1,z1],[x2,y2,z2],...]
    """
    xyz = np.atleast_2d(xyz)

    # calculate dec and map to 0-360 degree range
    dec = xy2dec(xyz[:, 0:2])
    mag = np.linalg.norm(xyz, axis=1)
    inc = z2inc(z=xyz[:, 2], mag=mag)

    return np.stack([dec, inc, mag]).T


def xy2dec(xy: ArrayLike) -> NDArray64:
    """calculates declination from x,y data

    The declination is defined as the angle between the negative
    y-axis and the projection of the vector onto the x-y plane.

    Parameters
    ----------
    xy: ndarray (2,n)
        x,y values of n vectors [[x1,y1],[x2,y2],...]

    Notes
    -----
    Declination is not defined for x = 0 and y = 0. This function returns 90° for these cases.
    """
    xy = np.atleast_2d(xy)
    x = xy[:, 0]
    y = xy[:, 1]
    return (90 + np.degrees(np.arctan2(y, x))) % 360


def z2inc(z: ArrayLike, mag: ArrayLike) -> NDArray64:
    """calculates inclination from z data

    Parameters
    ----------
    z: ndarray (n,)
        z values of n vectors [z1,z2,...]
    mag: ndarray (n,)
        magnitude of n vectors [m1,m2,...]
    """
    z_norm: NDArray64 = np.array(z) / np.array(mag)
    return -np.degrees(np.arcsin(z_norm))


def main():
    dim = np.array([[0, 0, 1], [0, 10, 1], [90, 45, 2]])
    xyz = dim2xyz(dim)
    print(xyz2dim(xyz))


if __name__ == "__main__":
    main()



import typing
from typing import Any, Tuple, Union

import logging

import numpy as np
from numpy.typing import ArrayLike, NDArray

from pypole import NDArray64, convert

LOG = logging.getLogger(__name__)


def get_random_sources(
    n_sources: int,
    x_range: tuple[float, float] = (-3.0e-6, 3.0e-6),
    y_range: tuple[float, float] = (-3.0e-6, 3.0e-6),
    z_range: tuple[float, float] = (1e-6, 4e-6),
    moment_range: tuple[float, float] = (1e-14, 1e-14),
) -> tuple[NDArray64, NDArray64]:
    """Generate diction of point source parameters

    Parameters
    ----------
    n_sources: int
             Number of sources per map
    x_range:  tuple (min, max)
        range for possible values
    y_range: tuple (min, max)
        range for possible values
    z_range: tuple (min, max)
        range of z for each source
    moment_range:  tuple (min, max)
        range of moments to generate

    Returns
    -------
        ndarrays of location and source_vector
            location (n_sources, 3): x_source, y_source, z_source
            source_vector (n_sources, 3): x,y,z components of dipole moment

    Notes
    -----
    The range of x,y positions for the dipole is currently limited to the central micron'
    """

    # get the locations of the sources
    locations = get_random_locations(n_sources, x_range, y_range, z_range)
    # reshape total location to be (n_maps, n_sources, (x,y,z))
    locations = locations.reshape(n_sources, 3)

    # calculate the x/y/z components
    dim = get_random_dim(n_sources, moment_range)
    xyz = convert.dim2xyz(dim)

    # reshape xyz to be (n_sources, (x,y,z))
    source_vector = xyz.reshape(n_sources, 3)

    return locations, source_vector


def get_random_dim(n_sources, moment_range):
    """Generate randomly distributed dipole moments on the unit sphere

    Parameters
    ----------
    moment_range: tuple (min, max)
        range of moments to generate
    n_maps: int
        number of maps to generate
    n_sources: int
        number of sources per map

    Returns
    -------
    dim: ndarray
        dipole moments in (declination, inclination, moment) format

    Examples
    --------
    >>> np.random.seed(0)
    >>> get_random_dim(2,(1e-14, 1e-14))
    >>> array([[ 1.97572861e+02, -1.18603363e+01,  1.00000000e-14], [ 2.57468172e+02, -5.15016644e+00,  1.00000000e-14]])

    """
    # get random declinations from uniform distribution (rand)
    declination = np.random.uniform(0, 360, n_sources)
    # get random inclinations from uniform distribution (rand)
    inclination = np.rad2deg(np.arccos(2 * np.random.rand(n_sources) - 1)) - 90
    # get uniform distribution of moment magnitude
    moment_scalar = np.random.uniform(moment_range[0], moment_range[1], size=n_sources)
    return np.stack([declination, inclination, moment_scalar]).T


def get_random_locations(n_sources, x_range, y_range, z_range):
    """Generate random locations for sources within the source region of the map (x_range, y_range, z_range)

    Parameters
    ----------
    n_sources: int
        number of sources per map
    x_range: tuple (min, max)
        range for possible values
    y_range: tuple (min, max)
        range for possible values
    z_range: tuple (min, max)
        range of z for each source

    Returns
    -------
    locations: ndarray
        x,y,z locations of sources

    Examples
    --------
    >>> np.random.seed(0)
    >>> get_random_locations(2, (-3e-6, 3e-6), (-3e-6, 3e-6), (1e-6, 4e-6))
    >>> array([[2.92881024e-07, 6.16580256e-07, 2.27096440e-06],[1.29113620e-06, 2.69299098e-07, 2.93768234e-06]])

    """

    # get uniform distribution of x/y locations
    x_source = np.random.uniform(x_range[0], x_range[1], size=n_sources)
    y_source = np.random.uniform(y_range[0], y_range[1], size=n_sources)
    z_source = np.random.uniform(z_range[0], z_range[1], size=n_sources)
    return np.stack([x_source, y_source, z_source]).T


def get_grid(
    pixels: Union[tuple[int, int], int] = (100, 100), pixel_size: float = 5e-6
) -> tuple[NDArray64, NDArray64]:
    """Generate observation coordinates of the map

    Parameters
    ----------
    n_pixel: int
        number of grid points (i.e. pixel) in the map
    pixel_size: float
        size of a single pixel (x/y are the same) in micron.
        defines: left map edge = -(pixel*pixel_size)/2, right map edge = (pixel*pixel_size)/2
    """

    if isinstance(pixels, int):
        LOG.warning(
            f"pixels should be a tuple of (x,y) pixel size. Setting to ({pixels},{pixels})"
        )
        pixels = (pixels, pixels)

    x_points = np.linspace(-pixels[0], pixels[0], pixels[0]) * pixel_size / 2
    y_points = np.linspace(-pixels[1], pixels[1], pixels[1]) * pixel_size / 2

    x_grid, y_grid = np.meshgrid(x_points, y_points)
    return x_grid.astype(np.float64), y_grid.astype(np.float64)



import typing

import numba
import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy.optimize import least_squares

from pypole import NDArray64, compute, maps
from pypole.dipole import dipole_field


def fit_dipole_n_maps(
    x_grid: NDArray[np.float64],
    y_grid: NDArray[np.float64],
    b_maps: NDArray[np.float64],
    initial_guess: NDArray[np.float64],
) -> NDArray[np.float64]:
    """fits a series of maps each with a single dipole.

    Parameters
    ----------
    x_grid : ndarray(pixel, pixel)
        x grid
    y_grid : ndarray(pixel, pixel)
        y grid
    b_maps : ndarray(n_maps, pixel, pixel)
        magnetic field map for all frames
    initial_guess : ndarray(n_maps, 3)
        initial guess for dipole parameters
    """
    n_maps = b_maps.shape[0]
    best_fit_dipoles = np.empty((n_maps, 6))

    for map_index in numba.prange(n_maps):
        best_fit_dipoles[map_index, :] = _fit_dipole(
            b_map=b_maps[map_index],
            p0=initial_guess[map_index],
            x_grid=x_grid,
            y_grid=y_grid,
        )
    return best_fit_dipoles


@numba.njit()
def __initial_guess_from_synthetic(mvec: NDArray[np.float64]) -> NDArray[np.float64]:
    """
    Get initial guess for dipolarity parameter calculation

    Parameters
    ----------
    mvec: ndarray(3,)
        magnetic vector

    Returns
    -------
    initial guess: ndarray(6,)
        initial guess for dipolarity parameter calculation
    """
    return np.array([0, 0, 5e-6, mvec[0], mvec[1], mvec[2]])


def dipole_residual(
    params: tuple[float, float, float, float, float, float],
    grid: NDArray[np.float64],
    data: NDArray[np.float64],
) -> NDArray[np.float64]:
    """residual function for fitting a single dipole to a magnetic field map

    Parameters
    ----------
    params: tuple(floats)
        parameters for field calculation:
            [x_source, y_source, z_source, mx, my, mz]
    grid: ndarray
        x,y grid
    data: ndarray
        field data

    Returns
    -------
        difference of calculated map from params and field data
    """
    x, y = grid
    arr = dipole_field(
        x_grid=x,
        y_grid=y,
        x_source=params[0],
        y_source=params[1],
        z_observed=params[2],
        mx=params[3],
        my=params[4],
        mz=params[5],
    )
    return arr.ravel() - data


def fit_dipole(b_map, p0, pixel_size=1):
    """fits a single dipole to a magnetic field map

    Parameters
    ----------
    b_map: ndarray
        magnetic field map in Tesla
    p0: tuple(floats)
        initial guess for dipole parameters
    pixel_size: float
        pixel size in meters

    Returns
    -------
    dipole parameters: tuple(floats)
        dipole parameters [x_source, y_source, z_source, mx, my, mz]
    """

    x_grid, y_grid = maps.get_grid(pixels=b_map.shape, pixel_size=pixel_size)
    return _fit_dipole(b_map, p0, x_grid, y_grid)


def _fit_dipole(b_map, p0, x_grid, y_grid):
    grid = np.vstack((y_grid.ravel(), x_grid.ravel()))
    return least_squares(
        dipole_residual,
        p0,
        args=(grid, b_map.T.ravel()),
        loss="huber",
        method="trf",
        gtol=2.3e-16,
        ftol=2.3e-16,
        xtol=2.3e-16,
        max_nfev=5000,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypole']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.6.2,<4.0.0',
 'numba>=0.56.4,<0.57.0',
 'numpy>=1.23.4,<2.0.0',
 'pre-commit>=2.20.0,<3.0.0',
 'rich>=10.14.0,<11.0.0',
 'scipy>=1.9.3,<2.0.0',
 'tqdm>=4.64.1,<5.0.0']

entry_points = \
{'console_scripts': ['pypole = pypole.__main__:app']}

setup_kwargs = {
    'name': 'pypole',
    'version': '0.1.1',
    'description': 'Awesome `pypole` is a Python cli/package created with https://github.com/TezRomacH/python-package-template',
    'long_description': '# pypole\n\n<div align="center">\n\n[![Build status](https://github.com/pypole/pypole/workflows/build/badge.svg?branch=master&event=push)](https://github.com/pypole/pypole/actions?query=workflow%3Abuild)\n[![Python Version](https://img.shields.io/pypi/pyversions/pypole.svg)](https://pypi.org/project/pypole/)\n[![Dependencies Status](https://img.shields.io/badge/dependencies-up%20to%20date-brightgreen.svg)](https://github.com/pypole/pypole/pulls?utf8=%E2%9C%93&q=is%3Apr%20author%3Aapp%2Fdependabot)\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Security: bandit](https://img.shields.io/badge/security-bandit-green.svg)](https://github.com/PyCQA/bandit)\n[![Pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pypole/pypole/blob/master/.pre-commit-config.yaml)\n[![Semantic Versions](https://img.shields.io/badge/%20%20%F0%9F%93%A6%F0%9F%9A%80-semantic--versions-e10079.svg)](https://github.com/pypole/pypole/releases)\n[![License](https://img.shields.io/github/license/pypole/pypole)](https://github.com/pypole/pypole/blob/master/LICENSE)\n![Coverage Report](assets/images/coverage.svg)\n\n`pypole` is a Python package for simple fitting of magnetic dipoles to magnetic maps measured with Quantum diamond microscope or SQUID microscope\n\n</div>\n\n## Very first steps\n\n\n### Poetry\n\nWant to know more about Poetry? Check [its documentation](https://python-poetry.org/docs/).\n\n<details>\n<summary>Details about Poetry</summary>\n<p>\n\nPoetry\'s [commands](https://python-poetry.org/docs/cli/#commands) are very intuitive and easy to learn, like:\n\n- `poetry add numpy@latest`\n- `poetry run pytest`\n- `poetry publish --build`\n\netc\n</p>\n</details>\n\n### Building and releasing your package\n\nBuilding a new version of the application contains steps:\n\n- Bump the version of your package `poetry version <version>`. You can pass the new version explicitly, or a rule such as `major`, `minor`, or `patch`. For more details, refer to the [Semantic Versions](https://semver.org/) standard.\n- Make a commit to `GitHub`.\n- Create a `GitHub release`.\n- And... publish 🙂 `poetry publish --build`\n\n## 🎯 What\'s next\n\nWell, that\'s up to you 💪🏻. I can only recommend the packages and articles that helped me.\n\n- [`Typer`](https://github.com/tiangolo/typer) is great for creating CLI applications.\n- [`Rich`](https://github.com/willmcgugan/rich) makes it easy to add beautiful formatting in the terminal.\n- [`Pydantic`](https://github.com/samuelcolvin/pydantic/) – data validation and settings management using Python type hinting.\n- [`Loguru`](https://github.com/Delgan/loguru) makes logging (stupidly) simple.\n- [`tqdm`](https://github.com/tqdm/tqdm) – fast, extensible progress bar for Python and CLI.\n- [`IceCream`](https://github.com/gruns/icecream) is a little library for sweet and creamy debugging.\n- [`orjson`](https://github.com/ijl/orjson) – ultra fast JSON parsing library.\n- [`Returns`](https://github.com/dry-python/returns) makes you function\'s output meaningful, typed, and safe!\n- [`Hydra`](https://github.com/facebookresearch/hydra) is a framework for elegantly configuring complex applications.\n- [`FastAPI`](https://github.com/tiangolo/fastapi) is a type-driven asynchronous web framework.\n\nArticles:\n\n- [Open Source Guides](https://opensource.guide/).\n- [A handy guide to financial support for open source](https://github.com/nayafia/lemonade-stand)\n- [GitHub Actions Documentation](https://help.github.com/en/actions).\n- Maybe you would like to add [gitmoji](https://gitmoji.carloscuesta.me/) to commit names. This is really funny. 😄\n\n## 🚀 Features\n\n### Development features\n\n- Supports for `Python 3.9` and higher.\n- [`Poetry`](https://python-poetry.org/) as the dependencies manager. See configuration in [`pyproject.toml`](https://github.com/pypole/pypole/blob/master/pyproject.toml) and [`setup.cfg`](https://github.com/pypole/pypole/blob/master/setup.cfg).\n- Automatic codestyle with [`black`](https://github.com/psf/black), [`isort`](https://github.com/timothycrosley/isort) and [`pyupgrade`](https://github.com/asottile/pyupgrade).\n- Ready-to-use [`pre-commit`](https://pre-commit.com/) hooks with code-formatting.\n- Type checks with [`mypy`](https://mypy.readthedocs.io); docstring checks with [`darglint`](https://github.com/terrencepreilly/darglint); security checks with [`safety`](https://github.com/pyupio/safety) and [`bandit`](https://github.com/PyCQA/bandit)\n- Testing with [`pytest`](https://docs.pytest.org/en/latest/).\n- Ready-to-use [`.editorconfig`](https://github.com/pypole/pypole/blob/master/.editorconfig), [`.dockerignore`](https://github.com/pypole/pypole/blob/master/.dockerignore), and [`.gitignore`](https://github.com/pypole/pypole/blob/master/.gitignore). You don\'t have to worry about those things.\n\n### Deployment features\n\n- `GitHub` integration: issue and pr templates.\n- `Github Actions` with predefined [build workflow](https://github.com/pypole/pypole/blob/master/.github/workflows/build.yml) as the default CI/CD.\n- Everything is already set up for security checks, codestyle checks, code formatting, testing, linting, docker builds, etc with [`Makefile`](https://github.com/pypole/pypole/blob/master/Makefile#L89). More details in [makefile-usage](#makefile-usage).\n- [Dockerfile](https://github.com/pypole/pypole/blob/master/docker/Dockerfile) for your package.\n- Always up-to-date dependencies with [`@dependabot`](https://dependabot.com/). You will only [enable it](https://docs.github.com/en/github/administering-a-repository/enabling-and-disabling-version-updates#enabling-github-dependabot-version-updates).\n- Automatic drafts of new releases with [`Release Drafter`](https://github.com/marketplace/actions/release-drafter). You may see the list of labels in [`release-drafter.yml`](https://github.com/pypole/pypole/blob/master/.github/release-drafter.yml). Works perfectly with [Semantic Versions](https://semver.org/) specification.\n\n### Open source community features\n\n- Ready-to-use [Pull Requests templates](https://github.com/pypole/pypole/blob/master/.github/PULL_REQUEST_TEMPLATE.md) and several [Issue templates](https://github.com/pypole/pypole/tree/master/.github/ISSUE_TEMPLATE).\n- Files such as: `LICENSE`, `CONTRIBUTING.md`, `CODE_OF_CONDUCT.md`, and `SECURITY.md` are generated automatically.\n- [`Stale bot`](https://github.com/apps/stale) that closes abandoned issues after a period of inactivity. (You will only [need to setup free plan](https://github.com/marketplace/stale)). Configuration is [here](https://github.com/pypole/pypole/blob/master/.github/.stale.yml).\n- [Semantic Versions](https://semver.org/) specification with [`Release Drafter`](https://github.com/marketplace/actions/release-drafter).\n\n## Installation\n\n```bash\npip install -U pypole\n```\n\nor install with `Poetry`\n\n```bash\npoetry add pypole\n```\n\n## 🛡 License\n\n[![License](https://img.shields.io/github/license/pypole/pypole)](https://github.com/pypole/pypole/blob/master/LICENSE)\n\nThis project is licensed under the terms of the `MIT` license. See [LICENSE](https://github.com/pypole/pypole/blob/master/LICENSE) for more details.\n\n## 📃 Citation\n\n```bibtex\n@misc{pypole,\n  author = {pypole},\n  title = {Awesome `pypole` is a Python cli/package created with https://github.com/TezRomacH/python-package-template},\n  year = {2022},\n  publisher = {GitHub},\n  journal = {GitHub repository},\n  howpublished = {\\url{https://github.com/pypole/pypole}}\n}\n```\n\n## Credits [![🚀 Your next Python package needs a bleeding-edge project structure.](https://img.shields.io/badge/python--package--template-%F0%9F%9A%80-brightgreen)](https://github.com/TezRomacH/python-package-template)\n\nThis project was generated with [`python-package-template`](https://github.com/TezRomacH/python-package-template)\n',
    'author': 'pypole',
    'author_email': 'michaelvolk1979@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/pypole/pypole',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



__version__ = "0.1.0a"

import logging
import os
import shutil
import sys
from pathlib import Path

PROJECT_PATH = Path(__file__).parent

SRC_PATH = PROJECT_PATH.parent
sys.path.append(str(SRC_PATH))

import numpy as np
from numpy.typing import NDArray

NDArray64 = NDArray[np.float64]
